@@ -88,11 +88,8 @@ sub create_call {
 ############################################################################
 # allocate new sockets for RTP
 #
-# Args: ($self,$callid,$cseq,$idfrom,$idto,$side,$addr,\@media)
-#   $callid: call-id
-#   $cseq:   sequence number for cseq
-#   $idfrom: ID for from-side
-#   $idto:   ID for to-side
+# Args: ($self,$nat_basic,$side,$addr,\@media)
+#   $nat_basic: [ callid, ...]
 #   $side:   0 if SDP is from request, else 1
 #   $addr:   IP where to create the new sockets
 #   \@media: media like returned from Net::SIP::SDP::get_media
@@ -107,7 +104,7 @@ sub create_call {
 ############################################################################
 sub allocate_sockets {
     my Net::SIP::NATHelper::Base $self = shift;
-    my $callid = shift;
+    my $callid = $_[0]->[0];
 
     my $call = $self->{calls}{$callid}
 	||= $self->create_call($callid);
@@ -117,11 +114,8 @@ sub allocate_sockets {
 
 ############################################################################
 # activate session
-# Args: ($self,$callid,$cseq,$idfrom,$idto;$param)
-#   $callid: call-id
-#   $cseq:   sequence number for cseq
-#   $idfrom: ID for from-side
-#   $idto:   ID for to-side
+# Args: ($self,$nat_basic;$param)
+#   $nat_basic: [ callid, ...]
 #   $param:  user defined param which gets returned from info_as_hash
 # Returns: ($info,$duplicate)
 #   $info:  hash from sessions info_as_hash
@@ -132,7 +126,7 @@ sub allocate_sockets {
 ############################################################################
 sub activate_session {
     my Net::SIP::NATHelper::Base $self = shift;
-    my $callid = shift;
+    my $callid = $_[0]->[0];
 
     my $call = $self->{calls}{$callid};
     unless ( $call ) {
@@ -144,11 +138,8 @@ sub activate_session {
 
 ############################################################################
 # close session(s)
-# Args: ($self,$callid,$cseq,$idfrom,$idto)
-#   $callid: call-id
-#   $cseq:   optional sequence number, only for CANCEL requests
-#   $idfrom: ID for from-side
-#   $idto:   ID for to-side
+# Args: ($self,$nat_basic)
+#   $nat_basic: [ callid, ...]
 # Returns: @session_info
 #   @session_info: list of hashes from session info_as_hash
 # Comment: this SIP packet should be forwarded, even if the call
@@ -157,7 +148,7 @@ sub activate_session {
 ############################################################################
 sub close_session {
     my Net::SIP::NATHelper::Base $self = shift;
-    my $callid = shift;
+    my $callid = $_[0]->[0];
 
     my $call = $self->{calls}{$callid};
     unless ( $call ) {
@@ -177,6 +168,7 @@ sub close_session {
 #       defaults to 3 minutes
 #    active: seconds for timeout of sockets used in sessions, defaults to
 #       30 seconds
+#    closed: seconds for timeout of sockets from closed sessions, default 1 sec
 # Returns: @expired
 #   @expired: list of infos about expired sessions using sessions info_as_hash
 ############################################################################
@@ -187,7 +179,8 @@ sub expire {
     $args{time}   ||= gettimeofday();
     $args{unused} ||= 3*60; # unused sockets after 3 minutes
     $args{active} ||= 30;   # active sessions after 30 seconds
-    DEBUG( 100,"expire now=$args{time} unused=$args{unused} active=$args{active}" );
+    $args{closed} ||= 1;    # closed sessions after 1 seconds
+    DEBUG( 100,"expire now=$args{time} unused=$args{unused} active=$args{active} closed=$args{closed}" );
     my @expired;
     my $calls = $self->{calls};
     foreach my $callid ( keys %$calls ) {
@@ -288,7 +281,6 @@ sub get_rtp_sockets {
 
     foreach my $m (@$media) {
 	my ($addr,$port,$range) = @{$m}{qw/addr port range/};
-
 	# allocate new sockets
 	my ($new_port,@socks) = create_rtp_sockets( $new_addr,$range );
 	unless (@socks) {
@@ -296,16 +288,20 @@ sub get_rtp_sockets {
 	    return;
 	}
 
-	# determine target for sock, e.g. original address
-	my $addr_bin = inet_aton($addr);
-	my @targets;
-	for( my $i=0;$i<@socks;$i++ ) {
-	    my $dst = sockaddr_in( $port+$i,$addr_bin );
-	    push @targets,$dst;
+	if (!$port or $addr eq '0.0.0.0' or $addr eq '::') {
+	    # RFC 3264 6.1 - stream marked as inactive
+	    DEBUG(50,"inactive stream" );
+	    push @new_media, [ $new_addr,0,\@socks,
+		# no target for socket on other side
+		[ map { undef } (0..$#socks) ]
+	    ];
+	} else {
+	    DEBUG( 100,"m_old=$addr $port/$range new_port=$new_port" );
+	    push @new_media, [ $new_addr,$new_port,\@socks,
+		# target for sock on other side is original address
+		[ map { ip_parts2sockaddr($addr,$port+$_) } (0..$#socks) ]
+	    ]
 	}
-
-	DEBUG( 100,"m_old=$addr $port/$range new_port=$new_port" );
-	push @new_media, [ $new_addr,$new_port,\@socks,\@targets ];
     }
 
     $self->{socket_count} += $need_sockets;
@@ -338,7 +334,7 @@ sub unget_rtp_sockets {
 ############################################################################
 
 package Net::SIP::NATHelper::Call;
-use fields qw( callid from );
+use fields qw( callid from expired );
 use Hash::Util 'lock_keys';
 use List::Util 'max';
 use Net::SIP::Debug;
@@ -356,12 +352,13 @@ sub new {
 
 ############################################################################
 # allocate sockets for rewriting SDP body
-# Args: ($nathelper,$self,$cseq,$idfrom,$idto,$side,$addr,$media)
+# Args: ($nathelper,$self,$nat_basic,$side,$addr,$media)
 # Returns: $media
 ############################################################################
 sub allocate_sockets {
     my Net::SIP::NATHelper::Call $self = shift;
-    my ($nathelper,$cseq,$idfrom,$idto,$side,$addr,$media) = @_;
+    my ($nathelper,$nat_basic,$side,$addr,$media) = @_;
+    my (undef,$cseq,$idfrom,$idto) = @$nat_basic;
 
     # find existing data for $idfrom,$cseq
     my $cseq_data = $self->{from}{$idfrom};
@@ -412,12 +409,13 @@ sub allocate_sockets {
 
 ############################################################################
 # activate session
-# Args: ($self,$cseq,$idfrom,$idto;$param)
+# Args: ($self,$nat_basic;$param)
 # Returns: ($info,$duplicate)
 ############################################################################
 sub activate_session {
     my Net::SIP::NATHelper::Call $self = shift;
-    my ($cseq,$idfrom,$idto,$param) = @_;
+    my ($nat_basic,$param) = @_;
+    my (undef,$cseq,$idfrom,$idto,$uri_from,$uri_to) = @$nat_basic;
 
     my $by_cseq = $self->{from}{$idfrom};
     my $data = $by_cseq && $by_cseq->{$cseq};
@@ -439,9 +437,17 @@ sub activate_session {
 	return;
     }
 
-    my $sess = $sessions->{$idto} = $self->create_session( $gfrom,$gto,$param );
+    my $sess = $sessions->{$idto} = $self->create_session( $gfrom,$gto,{
+	uri_from => $uri_from,
+	uri_to   => $uri_to,
+	%{$param || {}},
+    });
     DEBUG( 10,"new session {$sess->{id}} $self->{callid},$cseq $idfrom -> $idto" );
 
+    # expire the now unused previous sessions in this call immediately
+    # will be returned at the next call to Base::expire
+    $self->{expired} = [ $self->expire() ];
+
     return ( $sess->info_as_hash( $self->{callid},$cseq ), 0 );
 }
 
@@ -460,29 +466,36 @@ sub create_session {
 
 ############################################################################
 # close session
-# Args: ($self,$cseq,$idfrom,$idto)
-#   $cseq: optional sequence number, only for CANCEL requests
+# Args: ($self,$nat_basic)
+#   $nat_basic: [ callid, cseq, idfrom, idto ... ]
+#       cseq used only for CANCEL requests
 # Returns: @session_info
 #   @session_info: list of infos of all closed sessions, info is hash with
 #     callid,cseq,idfrom,idto,from,to,bytes_from,bytes_to
 ############################################################################
 sub close_session {
     my Net::SIP::NATHelper::Call $self = shift;
-    my ($cseq,$idfrom,$idto) = @_;
+    my (undef,$cseq,$idfrom,$idto) = @{ shift() };
 
     #DEBUG( 100,$self->dump );
 
     my @info;
     if ( $cseq ) {
-	# close initiated by CANCEL
+	# close initiated by CANCEL orr ACK to 401
 	my $data = $self->{from}{$idfrom};
 	$data = $data && $data->{$cseq};
-	my $sess = $data && delete( $data->{sessions}{$idto} ) or do {
+	if (my $sess = $data && delete( $data->{sessions}{$idto} )) {
+	    push @info, $sess->info_as_hash( $self->{callid},$cseq );
+	    DEBUG( 10,"close session {$sess->{id}} $self->{callid}|$cseq $idfrom -> $idto success" );
+	} else {
 	    DEBUG( 10,"tried to CANCEL non existing session in $self->{callid}|$cseq" );
-	    return;
-	};
-	push @info, $sess->info_as_hash( $self->{callid},$cseq );
-	DEBUG( 10,"close session {$sess->{id}} $self->{callid}|$cseq $idto,$idfrom success" );
+	}
+	if ($data && !%{$data->{sessions}}) {
+	    %{$data->{socket_groups_to}} = ();
+	    $data->{socket_group_from} = undef;
+	    DEBUG( 10,"cancel sessions $self->{callid}|$cseq $idfrom -> $idfrom - no more sessions" );
+	    delete $self->{from}{$idfrom}{$cseq};
+	}
 
     } else {
 	# close from BYE (which has different cseq then the INVITE)
@@ -492,17 +505,26 @@ sub close_session {
 	    my ($from,$to) = @$pair;
 	    my $by_cseq = $self->{from}{$from} || next;
 
-	    foreach my $cseq ( keys %$by_cseq ) {
-		my $sess = delete $by_cseq->{$cseq}{sessions}{$to} || next;
-		push @info, $sess->info_as_hash( $self->{callid},$cseq );
-		DEBUG( 10,"close session {$sess->{id}} $self->{callid}|$cseq $idto,$idfrom " );
+	    my @del_cseq;
+	    while (my ($cseq,$data) = each %$by_cseq) {
+		if (my $sess = delete $data->{sessions}{$to}) {
+		    push @info, $sess->info_as_hash( $self->{callid},$cseq );
+		    DEBUG( 10,"close session {$sess->{id}} $self->{callid}|$cseq $idfrom -> $idto " );
+		}
+		if (!%{$data->{sessions}}) {
+		    %{$data->{socket_groups_to}} = ();
+		    $data->{socket_group_from} = undef;
+		    DEBUG( 10,"bye sessions $self->{callid}|$cseq $idfrom -> $idto - no more sessions" );
+		    push @del_cseq, $cseq;
+		}
 	    }
+	    delete @{$by_cseq}{@del_cseq} if @del_cseq;
 	}
 	unless (@info) {
 	    DEBUG( 10,"tried to BYE non existing session in $self->{callid}" );
 	    return;
 	}
-	DEBUG( 10,"close sessions $self->{callid} $idto,$idfrom success" );
+	DEBUG( 10,"close sessions $self->{callid} $idfrom -> $idto success" );
     }
     return @info;
 }
@@ -519,10 +541,12 @@ sub expire {
     my Net::SIP::NATHelper::Call $self = shift;
     my %args = @_;
 
-    my $expire_unused = $args{time} - $args{unused};
-    my $expire_active = $args{time} - $args{active};
+    my $expire_unused = $args{unused} ? $args{time} - $args{unused} : 0;
+    my $expire_active = $args{active} ? $args{time} - $args{active} : 0;
+    my $expire_closed = $args{closed} ? $args{time} - $args{closed} : 0;
 
-    my @expired;
+    my @expired = @{$self->{expired} || []}; # from previous calls inside activate_session
+    @{$self->{expired}} = ();
     my %active_pairs; # mapping [idfrom,idto]|[idto,idfrom] -> session.created
     my $need_next_pass;
     my $by_from = $self->{from};
@@ -542,7 +566,7 @@ sub expire {
 		    my $sess = $sessions->{$idto};
 		    my $lastmod = max($sess->lastmod,$sess->{created});
 		    if ( $lastmod < $expire_active ) {
-			DEBUG( 10,"expired session {$sess->{id}} $cseq|$idfrom|$idto because lastmod($lastmod) < active($expire_active)" );
+			DEBUG( 10,"$self->{callid} expired session {$sess->{id}} $cseq|$idfrom|$idto because lastmod($lastmod) < active($expire_active)" );
 			my $sess = delete $sessions->{$idto};
 			push @expired, $sess->info_as_hash( $self->{callid}, $cseq, reason => 'expired' );
 
@@ -551,7 +575,7 @@ sub expire {
 			$active_pairs{ "$idto\0$idfrom" } || 0
 			) ) {
 			if ( $created > $sess->{created} ) {
-			    DEBUG( 10,"removed session {$sess->{id}} $cseq|$idfrom|$idto because there is newer session" );
+			    DEBUG( 10,"$self->{callid} removed session {$sess->{id}} $cseq|$idfrom|$idto because there is newer session" );
 			    my $sess = delete $sessions->{$idto};
 			    push @expired, $sess->info_as_hash( $self->{callid}, $cseq, reason => 'replaced' );
 			} elsif ( $created < $sess->{created} ) {
@@ -562,7 +586,7 @@ sub expire {
 			}
 		    } else {
 			# keep session
-			DEBUG( 100,"session {$sess->{id}} $idfrom -> $idto created=$sess->{created} stays active in pass#$pass" );
+			DEBUG( 100,"$self->{callid} session {$sess->{id}} $idfrom -> $idto created=$sess->{created} stays active in pass#$pass" );
 			$active_pairs{ "$idfrom\0$idto" } = $sess->{created};
 		    }
 		}
@@ -576,6 +600,7 @@ sub expire {
 		    $used{ $_->{sto} }++;
 		}
 
+
 		my $groups = $data->{socket_groups_to};
 		my %expired_sg;
 		my @v = values(%$groups);
@@ -586,11 +611,11 @@ sub expire {
 		    if ( ! $lastmod ) {
 			# was never used
 			if ( $v->{created} < $expire_unused ) {
-			    DEBUG( 10,"expired socketgroup $v->{id} because created($v->{created}) < unused($expire_unused)" );
+			    DEBUG( 10,"$self->{callid} expired socketgroup $v->{id} because created($v->{created}) < unused($expire_unused)" );
 			    $expired_sg{$v} = 1;
 			}
-		    } elsif ( $lastmod < $expire_active ) {
-			DEBUG( 10,"expired socketgroup $v->{id} because lastmod($lastmod) < active($expire_active)" );
+		    } elsif ( $lastmod < $expire_closed ) {
+			DEBUG( 10,"$self->{callid} expired socketgroup $v->{id} because lastmod($lastmod) < closed($expire_closed)" );
 			$expired_sg{$v} = 1;
 		    }
 		}
@@ -720,6 +745,7 @@ package Net::SIP::NATHelper::Session;
 use fields qw( sfrom sto created bytes_from bytes_to callbacks id param );
 use Net::SIP::Debug;
 use List::Util 'max';
+use Net::SIP::Util ':all';
 use Time::HiRes 'gettimeofday';
 
 # increased for each new session
@@ -838,7 +864,7 @@ sub callbacks {
 	my $recvaddr = $targets_to->[$i];
 	my $dstaddr = $targets_from->[$i];
 
-	push @cb, [
+	$dstaddr && push @cb, [
 	    $sockets_from->[$i],
 	    [
 		$fwd_data,
@@ -852,7 +878,7 @@ sub callbacks {
 	    ++$callback_id
 	];
 
-	push @cb, [
+	$recvaddr && push @cb, [
 	    $sockets_to->[$i],
 	    [
 		$fwd_data,
@@ -880,12 +906,7 @@ sub forward_data {
 	return;
     };
 
-    my $name = sub {
-	my $bin = shift;
-	use Socket;
-	my ($port,$addr) = unpack_sockaddr_in( $bin );
-	return inet_ntoa($addr).':'.$port;
-    };
+    my $name = sub { ip_sockaddr2string(shift) };
 
     if ( ! $$bytes ) {
 	if ( $peer eq $$rfrom ) {
